//============================================================================//
// This source file is part of work done by Reinder Nijhoff (reinder@infi.nl) //
// For the latest info, see http://developer.infi.nl                          //
//                                                                            //
// You're free to use the code in any way you like, modified, unmodified or   //
// cut'n'pasted into your own work.                                           //
//                                                                            //
// Part of this source is based on work by:                                   //
//    - Humus (http://esprit.campus.luth.se/~humus/)                          //
//    - Paul Baker (http://www.paulsprojects.net)                             //
//============================================================================//

#include <SDL/SDL.h>
#include <GL/gl.h>
#include <GL/glu.h>

#include "base_sdl.hpp"
#include "log.hpp"
#include "factory.hpp"


SDL_Surface *screen;

//==============================================================================
//  Register to Factory
//==============================================================================
bool fac_base_sdl  = Factory_base.reg<Base_SDL>( "sdl" );

//==============================================================================
//  Base_SDL::init() 
//==============================================================================
void Base_SDL::init() {
    if (SDL_Init(SDL_INIT_VIDEO) < 0) {
        log_error << "Error initialising SDL - " << SDL_GetError() << std::endl;
        exit(0);
    }    
}

//==============================================================================
//  Base::init()
//==============================================================================
void Base_SDL::init(Data_node* data) {    
    Base::init( data );    
    init();
}    

//==============================================================================
//  Base_SDL::close()
//==============================================================================
void Base_SDL::close() {	 
    if(!m_hardware_acc) {
	    SDL_SaveBMP( SDL_GetVideoSurface(), "screenshot.bmp" );
    }
}    

//==============================================================================
//  Base_SDL::~Base_SDL()
//==============================================================================
Base_SDL::~Base_SDL() {    
    SDL_Quit();
}

//==============================================================================
//  Base_SDL::init_window()
//==============================================================================
void Base_SDL::init_window( const int width, const int height, const bool fullscreen, const bool hardware_acc ) {
    m_hardware_acc = hardware_acc;
    m_width = width;
    m_height = height;
    m_fullscreen = fullscreen;
    
    // inform the user 
    log_function << "Window initialization" << std::endl;    
    
    const SDL_VideoInfo* m_VideoInfo;
    int rgb_size[3];
    bool fsaa = false;
    int value;

    // Retrieve current video info
    if ((m_VideoInfo = SDL_GetVideoInfo()) == 0) {
        log_error << "Video query failed - " << SDL_GetError() << std::endl;
        exit(0);
    }
        
    Uint32 full = m_fullscreen ? SDL_FULLSCREEN : 0;

    /* initialize the display */
fprintf(stderr, "bpp %d\n", m_VideoInfo->vfmt->BitsPerPixel);
    switch (m_VideoInfo->vfmt->BitsPerPixel) {
        case 8:
        rgb_size[0] = 3;
        rgb_size[1] = 3;
        rgb_size[2] = 2;
        break;
        case 15:
        case 16:
        rgb_size[0] = 5;
        rgb_size[1] = 5;
        rgb_size[2] = 5;
        break;
            default:
        rgb_size[0] = 8;
        rgb_size[1] = 8;
        rgb_size[2] = 8;
        break;
    }
    if(m_hardware_acc) {
        SDL_GL_SetAttribute( SDL_GL_RED_SIZE, rgb_size[0] );
        SDL_GL_SetAttribute( SDL_GL_GREEN_SIZE, rgb_size[1] );
        SDL_GL_SetAttribute( SDL_GL_BLUE_SIZE, rgb_size[2] );
//        SDL_GL_SetAttribute( SDL_GL_DEPTH_SIZE, m_VideoInfo->vfmt->BitsPerPixel );
        SDL_GL_SetAttribute( SDL_GL_DEPTH_SIZE, 24 );
        SDL_GL_SetAttribute( SDL_GL_DOUBLEBUFFER, 1 );
    }    
/*    if ( fsaa ) {
        SDL_GL_SetAttribute( SDL_GL_MULTISAMPLEBUFFERS, 1 );
        SDL_GL_SetAttribute( SDL_GL_MULTISAMPLESAMPLES, fsaa );
    }*/

    // Set the video mode
    if (!SDL_SetVideoMode(m_width, m_height, m_VideoInfo->vfmt->BitsPerPixel, (m_hardware_acc?SDL_OPENGL:0) | (m_fullscreen?SDL_FULLSCREEN:0))) {
        log_error << "Video mode set failed - " << SDL_GetError() << std::endl;
        exit(0);
    }
    
    screen = SDL_GetVideoSurface();
    m_width = screen->w;
    m_height = screen->h;     
    
    if(m_hardware_acc) {
        SDL_GL_GetAttribute( SDL_GL_BUFFER_SIZE, &value );
        log_debug << "Screen BPP : " <<  value << std::endl;
        log_debug << "Vendor     : " <<  glGetString( GL_VENDOR ) << std::endl;
        log_debug << "Renderer   : " <<  glGetString( GL_RENDERER ) << std::endl;
        log_debug << "Version    : " <<  glGetString( GL_VERSION ) << std::endl;
    }    
//    log_debug << "Extensions : " << glGetString( GL_EXTENSIONS ) << std::endl;

    log_debug << "Size       : " <<  get_window_width() << " x " << get_window_height() << std::endl;    
    log_debug << "Fullscreen : " <<  get_fullscreen() << std::endl;    
     
       if(m_hardware_acc) {    
           SDL_GL_GetAttribute( SDL_GL_RED_SIZE, &value );
           log_debug << "SDL_GL_RED_SIZE    : requested " <<  rgb_size[0] << ", got " << value << std::endl;
           SDL_GL_GetAttribute( SDL_GL_GREEN_SIZE, &value );
           log_debug << "SDL_GL_GREEN_SIZE  : requested " <<  rgb_size[1] << ", got " << value << std::endl;
           SDL_GL_GetAttribute( SDL_GL_BLUE_SIZE, &value );
           log_debug << "SDL_GL_BLUE_SIZE   : requested " <<  rgb_size[2] << ", got " << value << std::endl;
           SDL_GL_GetAttribute( SDL_GL_DEPTH_SIZE, &value );
           log_debug << "SDL_GL_DEPTH_SIZE  : requested " << m_VideoInfo->vfmt->BitsPerPixel << ", got " << value << std::endl;
           SDL_GL_GetAttribute( SDL_GL_DOUBLEBUFFER, &value );
           log_debug << "SDL_GL_DOUBLEBUFFER: requested 1, got " << value << std::endl;
    }        
/*    if ( fsaa ) {
        SDL_GL_GetAttribute( SDL_GL_MULTISAMPLEBUFFERS, &value );
        log_debug << "SDL_GL_MULTISAMPLEBUFFERS: requested 1, got " << value << std::endl;
        SDL_GL_GetAttribute( SDL_GL_MULTISAMPLESAMPLES, &value );
        log_debug << "SDL_GL_MULTISAMPLESAMPLES: requested " << fsaa << ", got " << value << std::endl;
    }*/
    set_caption( m_caption.c_str() );
    set_mouse_cursor_visibility( !m_hide_cursor );
    
    log_success << std::endl;
}

//==============================================================================
//  Base_SDL::swap_buffers()
//==============================================================================
void Base_SDL::swap_buffers(void) {
    if(m_hardware_acc) {
        glFlush();
        SDL_GL_SwapBuffers();    
    } else {
        SDL_UpdateRect(screen, 0, 0, get_window_width(), get_window_height() );
       }    
}

//==============================================================================
//  Base_SDL::set_caption()
//==============================================================================
void Base_SDL::set_caption(const char *title) {
    SDL_WM_SetCaption(title, ""); 
}

//==============================================================================
//  Base_SDL::get_sys_time()
//==============================================================================
float Base_SDL::get_sys_time() const {
    return SDL_GetTicks() * 0.001f;    
}

//==============================================================
// Base_SDL::update()
//==============================================================
bool Base_SDL::update() {
    // Timer
    update_timer();    
    // Backup current frame keys
    memcpy(m_keys_last_frame, m_keys, sizeof(m_keys));
    
    SDL_Event event;
    
    // Retrieve all events from the queue
    while( SDL_PollEvent(&event) ) {
        switch (event.type) {
                // Handle key presses
                case (SDL_KEYDOWN): m_keys[event.key.keysym.sym] = true;    
                                    break;
                case (SDL_KEYUP):   m_keys[event.key.keysym.sym] = false;    
                                    break;        
                  // Todo: handle program exit
                case (SDL_QUIT):    return false;
                                    break;
        }
    }  

    // Set the key released states
    for (int i = 0; i < SDLK_LAST; i++) {
        m_keys_released[i] = (m_keys_last_frame[i] ^ m_keys[i]) & m_keys_last_frame[i];
    }
    // Get the current mouse state
    m_mouse_buttons = SDL_GetMouseState(&m_mouse_x, &m_mouse_y);
        
    /* default key-functions */
    if (m_keys[SDLK_ESCAPE] || m_keys[SDLK_q]) {
        return false;
    }
    
    return true;
}

//==============================================================
//  Base_SDL::set_mouse_position()
//==============================================================
void Base_SDL::set_mouse_position(int x, int y) {
    SDL_WarpMouse(x, y);
    m_mouse_x = x;
    m_mouse_y = y;
}

//==============================================================
// Base_SDL::set_mouse_cursor_visibility()
//==============================================================
void Base_SDL::set_mouse_cursor_visibility(bool visible) {
    if( visible ) {
        SDL_ShowCursor( SDL_ENABLE );
    } else {
        SDL_ShowCursor( SDL_DISABLE );
        set_mouse_position( get_window_width()/2, get_window_height()/2 );
    } 
}    
    
//==============================================================
// Base_SDL::get_buffer()
//==============================================================
unsigned char* Base_SDL::get_buffer() {
    return (unsigned char *) screen->pixels;
}
    
//==============================================================
// Base_SDL::putpixel()
//==============================================================
void Base_SDL::put_pixel(int x, int y, Color color) {
    int bpp = screen->format->BytesPerPixel;
    /* Here p is the address to the pixel we want to set */
    Uint8 *p = (Uint8 *)screen->pixels + y * screen->pitch + x * bpp;

    int r = (int)(255.0f * color.x);
    int g = (int)(255.0f * color.y);
    int b = (int)(255.0f * color.z); 

/*           
    if(r<0) r = 0;
    if(g<0) g = 0;
    if(b<0) b = 0;        
    
    if(r>255) r = 255;
    if(g>255) g = 255;
    if(b>255) b = 255;        
*/
 
    Uint32 pixel = ((r & 0xff) << 16) + ((g & 0xff) << 8) + (b & 0xff);
//    Uint32 pixel = ((r) << 16) + ((g) << 8) + (b);
    
    switch(bpp) {
        case 4: *(Uint32 *)p = pixel;
                break;				
        case 1: *p = pixel;
                break;
        case 2:
                *(Uint16 *)p = pixel;
                break;
        case 3: if(SDL_BYTEORDER == SDL_BIG_ENDIAN) {
                    p[0] = (pixel >> 16) & 0xff;
                    p[1] = (pixel >> 8) & 0xff;
                    p[2] = pixel & 0xff;
                } else {
                    p[0] = pixel & 0xff;
                    p[1] = (pixel >> 8) & 0xff;
                    p[2] = (pixel >> 16) & 0xff;
                }
                break;    
    }
}
