//============================================================================//
// This source file is part of work done by Reinder Nijhoff (reinder@infi.nl) //
// For the latest info, see http://developer.infi.nl                          //
//                                                                            //
// You're free to use the code in any way you like, modified, unmodified or   //
// cut'n'pasted into your own work.                                           //
//                                                                            //
// Part of this source is based on work by:                                   //
//    - Humus (http://esprit.campus.luth.se/~humus/)                          //
//    - Paul Baker (http://www.paulsprojects.net)                             //
//============================================================================//

#ifndef _BATCH_
#define _BATCH_ 1

#include <string>
#include <vector>

#include "includes.hpp"
#include "vector.hpp"

class Data_node;
class Collision_triangle;

//==============================================================================
//  Simplified batch, only use floats for data and unsigned int for indices
//==============================================================================

enum Primitive_type {
    PRIM_TRIANGLES      = 0,
    PRIM_QUADS          = 1,
    PRIM_TRIANGLE_STRIP = 2,
    PRIM_POINTS         = 3
};

enum Attribute_type {
    ATT_VERTEX    = 0,
    ATT_NORMAL    = 1,
    ATT_TEXCOORD  = 2,
    ATT_COLOR     = 3,
    ATT_TANGENT_S = 4,
    ATT_TANGENT_T = 5
};

struct Format {
    Attribute_type att_type;
    unsigned int size;
    unsigned int offset;
    unsigned int index;
};

//==============================================================================
//  Batch
//==============================================================================
class Batch {
public:
    Batch();
    virtual ~Batch();
    
    virtual void init(Data_node* data);

    void clean();

    unsigned int get_index(const unsigned int index) const {
        return m_indices[index];
    }

    void set_index(const unsigned int index, const unsigned int value) {
        m_indices[index] = value;
    }

    float *get_vertices() const                             { return m_vertices; }
    unsigned short *get_indices() const                      { return m_indices;  }
    const unsigned int get_vertex_count() const             { return m_num_vertices;  }
    const unsigned int get_index_count()  const             { return m_num_indices;   }
    const unsigned int get_vertex_size()  const             { return m_vertex_size; }

    const unsigned int get_format_count() const             { return m_formats.size(); }
    const Format &get_format(unsigned int index) const      { return m_formats[index]; }

    const Primitive_type get_primitive_type() const         { return m_primitive_type; }
    void set_primitive_type(const Primitive_type prim_type) { m_primitive_type = prim_type; }

    void set_vertices(void *vertex_array, const unsigned int vertex_count, const unsigned int size) {
        m_vertices = (float *) vertex_array;
        m_num_vertices = vertex_count;
        m_vertex_size = size;
        m_vertex_elements = m_vertex_size / sizeof( float );
    }
    void set_indices(unsigned short *index_array, const unsigned int index_count ) {
        m_indices = index_array;
        m_num_indices = index_count;
    }

    void add_format(const Attribute_type att_type, const unsigned int size, const unsigned int offset, const unsigned int index = 0) {
        Format format;

        format.att_type = att_type;
        format.size = size;
        format.offset = offset;
        format.index = index;

        m_formats.push_back(format);

		if(index == 0) {
			switch( att_type ) {
			   case ATT_VERTEX:     m_vertex_offset = offset;
			                        break;
			   case ATT_NORMAL:     m_normal_offset = offset;
			                        break;
			   case ATT_TEXCOORD:   m_texcoord_offset = offset;
			                        break;					
			   case ATT_TANGENT_S:  m_tangent_s_offset = offset;
			                        break;											                    
			   case ATT_TANGENT_T:  m_tangent_t_offset = offset;
			                        break;											                    						                    
			}	 
	    }
        
        if( size+offset > m_vertex_elements ) {
			m_vertex_elements = size+offset;
			m_vertex_size = m_vertex_elements * sizeof(float);
		}
    }
    
    bool find_attribute(const Attribute_type att_type, const unsigned int index = 0, unsigned int *where = NULL) const;
    bool insert_attribute(const Attribute_type att_type, const unsigned int size, const unsigned int index = 0);
    bool remove_attribute(const Attribute_type att_type, const unsigned int index = 0);

    virtual void render() {}
    virtual void render_solid( int mask ) {}
    
    bool add_normals();
    bool add_tangent_space();
    
    virtual bool upload_batch() {};
       
    void read_from_file_bin(const std::string& name);
    void write_to_file_bin(const std::string& name);
    void read_from_file_mesh(const std::string& name);
    void write_to_file_mesh(const std::string& name);
    
    float get_value(const float *src, const unsigned int index) {
        return *(((float *) src) + index);
    }
    
    inline Vector3f& get_vertex_by_index( unsigned int index )      { return get_vertex   ( m_indices[index] ); }
    inline Vector3f& get_normal_by_index( unsigned int index )      { return get_normal   ( m_indices[index] ); }
    inline Vector2f& get_texcoord_by_index( unsigned int index )    { return get_texcoord ( m_indices[index] ); }
    inline Vector3f& get_tangent_s_by_index( unsigned int index )   { return get_tangent_s( m_indices[index] ); }
    inline Vector3f& get_tangent_t_by_index( unsigned int index )   { return get_tangent_t( m_indices[index] ); }

    inline Vector3f& get_vertex( unsigned int index )      { return *((Vector3f *)(&m_vertices[ index*m_vertex_elements + m_vertex_offset ])); }
    inline Vector3f& get_normal( unsigned int index )      { return *((Vector3f *)(&m_vertices[ index*m_vertex_elements + m_normal_offset ])); }
    inline Vector2f& get_texcoord( unsigned int index )    { return *((Vector2f *)(&m_vertices[ index*m_vertex_elements + m_texcoord_offset ])); }
    inline Vector3f& get_tangent_s( unsigned int index )   { return *((Vector3f *)(&m_vertices[ index*m_vertex_elements + m_tangent_s_offset ])); }
    inline Vector3f& get_tangent_t( unsigned int index )   { return *((Vector3f *)(&m_vertices[ index*m_vertex_elements + m_tangent_t_offset ])); }    


protected:
    float *m_vertices;
    unsigned short *m_indices;

    unsigned short m_num_vertices, m_num_indices;
    unsigned short m_vertex_size, m_vertex_elements;
    unsigned short m_vertex_offset, m_normal_offset, m_texcoord_offset, m_tangent_s_offset, m_tangent_t_offset;

    std::vector<Format> m_formats;
    Primitive_type m_primitive_type;    
};

#endif /* _BATCH_ */
