//============================================================================//
// This source file is part of work done by Reinder Nijhoff (reinder@infi.nl) //
// For the latest info, see http://developer.infi.nl                          //
//                                                                            //
// You're free to use the code in any way you like, modified, unmodified or   //
// cut'n'pasted into your own work.                                           //
//                                                                            //
// Part of this source is based on work by:                                   //
//    - Humus (http://esprit.campus.luth.se/~humus/)                          //
//    - Paul Baker (http://www.paulsprojects.net)                             //
//============================================================================//

#ifndef _CAMERA_
#define _CAMERA_ 1

#include "includes.hpp"
#include "vector.hpp"
#include "geometry.hpp"
#include "controller.hpp"
#include "geometry.hpp"

class Data_node;
class Ray;

enum Camera_type {
    CAM_3D                = 0,
    CAM_2D                = 1
};

enum Frustum_plane {
    FRUSTUM_FRONT        = 0,
    FRUSTUM_LEFT        = 1,
    FRUSTUM_TOP            = 2,
    FRUSTUM_RIGHT        = 3,
    FRUSTUM_BOTTOM        = 4,
    FRUSTUM_BACK        = 5,
    FRUSTUM_PLANES        = 6
};    

//==============================================================================
//  Camera
//==============================================================================
class Camera : public Controllable_object {
public:
    Camera();
    ~Camera() {}
    virtual void init(Data_node* data);

    void update();
    
    void set_type(const Camera_type type) { m_type = type; }
    Camera_type get_type() const { return m_type; }
    
    void set_perspective(float fov_y, float near_plane, float far_plane);
    void get_perspective(float &fov_y, float &near_plane, float &far_plane);
    
    void  set_exposure(float exposure) { m_exposure = exposure; }
    float get_exposure() { return m_exposure; }
     
    void update_frustum();
    Plane& get_frustum_plane(int index) const { return (Plane &)m_frustum_planes[index]; }
        
    const Matrix4x4f& get_modelview_matrix()  { return m_modelview_matrix;  }
    const Matrix4x4f& get_projection_matrix() { return m_projection_matrix; }
    
    // unproject
    inline bool project(const Vector3f &point, int &x, int &y) {
        return project( Vector4f( point.x, point.y, point.z, 1.0f ), x, y);
    }    
    inline bool project(const Vector4f &point, int &x, int &y) {
        Vector2f proj;
        project( point, proj );
        
        x = (int) proj.x;
        y = (int) proj.y;
        
        return true;
    }
    inline bool project(const Vector3f &point, Vector2f &proj) {
        return project( Vector4f( point.x, point.y, point.z, 1.0f ), proj);
    }    
    inline bool project(Vector4f point, Vector2f &proj) {
        Vector4f vec = m_projection_modelview_matrix * point;
        
        if( vec.w == 0) {
            proj = Vector2f( 0.0f, 0.0f );
            return false;
        }       
        
        float factor = (0.5f/vec.w);
        vec.x *= factor;
        vec.y *= factor;
        
        proj.x = ((vec.x + 0.5f) * m_viewport_width);
        proj.y = ((vec.y + 0.5f) * m_viewport_height);
        
        return true;
    }    
    inline bool project( AA_box* aa_box, int &l, int &r, int &t, int &b) {
		if( aa_box->intersect( get_position() ) ) return false;   

		const Vector3f centre = aa_box->get_centre();
		const Vector3f diagonal = aa_box->get_diagonal();	
		   
		int px, py;
		project( centre+Vector3f( diagonal.x, diagonal.y, diagonal.z ), px, py );
		l=px; r=px; t=py; b=py;
		project( centre+Vector3f(-diagonal.x, diagonal.y, diagonal.z ), px, py );
		l=get_min(px, l); r=get_max(px, r); t=get_min(py, t); b=get_max(py, b);
		project( centre+Vector3f( diagonal.x,-diagonal.y, diagonal.z ), px, py );
		l=get_min(px, l); r=get_max(px, r); t=get_min(py, t); b=get_max(py, b);
		project( centre+Vector3f( diagonal.x, diagonal.y,-diagonal.z ), px, py );						
		l=get_min(px, l); r=get_max(px, r); t=get_min(py, t); b=get_max(py, b);
		project( centre+Vector3f(-diagonal.x,-diagonal.y, diagonal.z ), px, py );
		l=get_min(px, l); r=get_max(px, r); t=get_min(py, t); b=get_max(py, b);
		project( centre+Vector3f( diagonal.x,-diagonal.y,-diagonal.z ), px, py );
		l=get_min(px, l); r=get_max(px, r); t=get_min(py, t); b=get_max(py, b);
		project( centre+Vector3f(-diagonal.x, diagonal.y,-diagonal.z ), px, py );
		l=get_min(px, l); r=get_max(px, r); t=get_min(py, t); b=get_max(py, b);			
		project( centre+Vector3f(-diagonal.x,-diagonal.y,-diagonal.z ), px, py );
		l=get_min(px, l); r=get_max(px, r); t=get_min(py, t); b=get_max(py, b);
		
		return true;
	}
    
    
    inline void get_ray( Ray& ray, const Vector2f& coord ) {
        Vector3f direction = (m_direction + coord.x*m_dx + coord.y*m_dy).get_normalized();
        ray.set( get_position(), direction );
    }
    
    bool intersect( AA_box* aa_box ) {
	    float m, n; 		
		for(int i=0; i<6; ++i) { 
			Plane p = m_frustum_planes[i];

			m = p.distance( aa_box->get_centre() );
			const Vector3f& normal = p.get_normal();
			n = dot( Vector3f( fabs(normal.x), fabs(normal.y), fabs(normal.z)), aa_box->get_diagonal() );
			if (m + n < 0) return false;
		}
	    return true;
	}
    bool intersect( Sphere* sphere ) {
	    float distance;
		for(int i=0; i<6; ++i) { 
			Plane p = m_frustum_planes[i];

			distance = p.distance( sphere->get_centre() );
			if(distance < -sphere->get_radius() ) {
			    return false;
			}
		}
	    return true;
	}
private:
    float m_near_plane;
    float m_far_plane;
    float m_fov_y;
    float m_exposure;
    
    Camera_type m_type;
    Plane m_frustum_planes[FRUSTUM_PLANES];
        
    Matrix4x4f m_modelview_matrix, m_projection_matrix, m_projection_modelview_matrix;
    Vector3f m_direction, m_dx, m_dy;
    float m_viewport_width, m_viewport_height;
};

#endif /* _CAMERA_ */

