//============================================================================//
// This source file is part of work done by Reinder Nijhoff (reinder@infi.nl) //
// For the latest info, see http://developer.infi.nl                          //
//                                                                            //
// You're free to use the code in any way you like, modified, unmodified or   //
// cut'n'pasted into your own work.                                           //
//                                                                            //
// Part of this source is based on work by:                                   //
//    - Humus (http://esprit.campus.luth.se/~humus/)                          //
//    - Paul Baker (http://www.paulsprojects.net)                             //
//============================================================================//

#include "controller.hpp"
#include "data.hpp"
#include "base.hpp"

#include "factory.hpp"

bool fac_con_def  = Factory_controller.reg<Controller>( "controller" );
bool fac_con_fps  = Factory_controller.reg<Controller_fps>( "fps" );
bool fac_con_osc  = Factory_controller.reg<Controller_oscillator>( "oscillator" );

//==============================================================================
//  Controllable_object::init(Data_node* data);
//==============================================================================
void Controllable_object::init(Data_node* data) {
    get_position() = Vector3f(0.0, 0.0, 0.0);
    get_orientation().load_identity();

    if( Data_value* value = data->get_value( "controller" ) ) {
		m_controller = Factory_controller.create( value );
	}
    data->get_value( "position", get_position() );
    data->get_value( "orientation", (Matrix3x3f &)get_orientation() );
}    
 
//==============================================================================
//  Controllable_object::Controllable_object()
//==============================================================================
Controllable_object::Controllable_object() { 
    m_default_controller = Factory_controller.create( "controller" );
    m_controller = m_default_controller;
}

//==============================================================================
//  Controllable_object::first_update()
//==============================================================================
void Controllable_object::first_update() {
	 m_controller->first_update();
}
 
//==============================================================================
//  Controllable_object::~Controllable_object()
//==============================================================================
Controllable_object::~Controllable_object() {              
}

//==============================================================================
//  Controller::init()
//==============================================================================
void Controller::init(Data_node* data) {
	 m_orientation.load_identity();
}

//==============================================================================
//  Controller_fps::Controller_fps()
//  remember, z is up
//==============================================================================
Controller_fps::Controller_fps() {
    m_speed = 15.0f;
}

//==============================================================================
//  Controller_fps::init(Data_node* data);
//==============================================================================
void Controller_fps::init(Data_node* data) {
    Controller::init( data );
     
    data->get_value( "speed", m_speed );
}
      
//==============================================================================
//  Controller_fps::set_rotation()
//==============================================================================
void Controller_fps::set_rotation(float angle_z, float angle_y) {
    Vector3f axis = cross( m_center - m_position, m_up);
    axis.normalize();
        
    set_rotation(angle_y, axis);
    set_rotation(angle_z, Vector3f(0, 0, 1) );
}

//==============================================================================
//  Controller_fps::set_rotation()
//==============================================================================
void Controller_fps::set_rotation(float angle, Vector3f axis) {
    Vector3f view = m_center - m_position;
    Matrix3x3f rotation;

    rotation.set_rotation_axis( angle, axis );
	
    view = rotation * view;
    m_up = rotation * m_up;
    m_up.normalize();    
    
	m_center = m_position + view;

    m_strafe = cross(view.get_normalized(), m_up).get_normalized();
   
    m_orientation.look_at( m_position, m_center, m_up);
    
	m_activity = ACTIVE;
}

//==============================================================================
//  Controller_fps::strafe()
//==============================================================================
void Controller_fps::strafe(float speed) {   
    m_position += m_strafe * speed;
    m_center   += m_strafe * speed;

    m_activity = ACTIVE;            
}

//==============================================================================
//  Controller_fps::move()
//==============================================================================
void Controller_fps::move(float speed) {
    Vector3f movement = m_center - m_position;
    movement.normalize();

    movement *= speed;
 
    m_position += movement;
    m_center   += movement;
	    
    m_activity = ACTIVE;            
}

//==============================================================================
//  Controller_fps::update()
//==============================================================================
void Controller_fps::update(void) {	 
	m_activity = FROZEN;
	 
    int middle_x = base->get_window_width() / 2;
    int middle_y = base->get_window_height() / 2;
    
    int mouse_x = base->get_mouse_x();
    int mouse_y = base->get_mouse_y();
    
    float angle_y = 0.0f;                    
    float angle_z = 0.0f;                    
    
    float speed = m_speed * base->get_frame_interval();

    if( base->is_key_pressed(KEY_UP) || base->is_key_pressed(KEY_W) ) {
        move(speed);
    }
    if( base->is_key_pressed(KEY_DOWN) || base->is_key_pressed(KEY_S) ) {            
        move(-speed);
    }
    if( base->is_key_pressed(KEY_LEFT) || base->is_key_pressed(KEY_A) ) {            
        strafe(-speed);
    }
    if( base->is_key_pressed(KEY_RIGHT) || base->is_key_pressed(KEY_D) ) {           
        strafe(speed);
    }

    // If our cursor is not in the middle
    if((mouse_x != middle_x) || (mouse_y != middle_y) ) {
        base->set_mouse_position(middle_x,middle_y);

        angle_y = (float)( (middle_x - mouse_x) ) / 500.0f;        
        angle_z = (float)( (middle_y - mouse_y) ) / 500.0f;

        set_rotation(angle_y, angle_z);
    }    
}

//==============================================================================
//  Controller_fps::first_update()
//==============================================================================
void Controller_fps::first_update() {
	Vector3f view = m_center - m_position;
    m_strafe = cross(view.get_normalized(), m_up).get_normalized();
    
    int middle_x = base->get_window_width() / 2;
    int middle_y = base->get_window_height() / 2;
    
    base->set_mouse_position(middle_x,middle_y);
}

//==============================================================================
//  Controller_oscillator::Controller_oscillator()
//==============================================================================
Controller_oscillator::Controller_oscillator() {
    m_frequency = m_amplitude = Vector3f( 1.0f, 1.0f, 1.0f );
}

//==============================================================================
//  Controller_oscillator::init()
//==============================================================================
void Controller_oscillator::init(Data_node* data) {
    Controller::init( data );
     
    data->get_value( "amplitude", m_amplitude );
    data->get_value( "frequency", m_frequency );    
}

//==============================================================================
//  Controller_oscillator::update()
//==============================================================================
void Controller_oscillator::update(void) {
	m_activity = ACTIVE;
	
	float time = base->get_sys_time();
	m_position = m_origin;
	m_position.x += sinf( time * m_frequency.x ) * m_amplitude.x;
	m_position.y += sinf( time * m_frequency.y ) * m_amplitude.y;
	m_position.z += sinf( time * m_frequency.z ) * m_amplitude.z;
}
