//============================================================================//
// This source file is part of work done by Reinder Nijhoff (reinder@infi.nl) //
// For the latest info, see http://developer.infi.nl                          //
//                                                                            //
// You're free to use the code in any way you like, modified, unmodified or   //
// cut'n'pasted into your own work.                                           //
//                                                                            //
// Part of this source is based on work by:                                   //
//    - Humus (http://esprit.campus.luth.se/~humus/)                          //
//    - Paul Baker (http://www.paulsprojects.net)                             //
//============================================================================//

#ifndef _GEOMETRY_
#define _GEOMETRY_ 1

#include <iostream>

#include "includes.hpp"
#include "vector.hpp"

//==============================================================================
//  Ray
//==============================================================================
class Ray
{
public:
    Ray() {};
    Ray( Vector3f& origin, Vector3f& direction ) : m_origin(origin), m_direction(direction) {}
      
    void set_origin( Vector3f& origin )            { m_origin = origin; }
    void set_direction( Vector3f& direction )      { m_direction = direction; }
    void set( Vector3f origin, Vector3f direction ) {
         m_origin = origin; m_direction = direction;
    }
     
    inline const Vector3f& get_origin() { return m_origin; }
    inline const Vector3f& get_direction() { return m_direction; }
    
    enum Hit {
        MISS   =  0,
        HIT    =  1,
        INPRIM = -1
    };   
protected:
    Vector3f m_origin;
    Vector3f m_direction;
};

//==============================================================================
//  Plane
//==============================================================================
class Plane {
public:
    // constructors 
    Plane(void) {}
    Plane(const Vector3f& n_normal, float n_d ) : m_normal(n_normal), m_d(n_d) {}    
    
    void set_normal( Vector3f normal )        { m_normal = normal; }
    void set_distance( float distance )     { m_d      = distance; }
    
    Vector3f& get_normal()   { return m_normal; }
    float&    get_distance() { return m_d; }
        
    int intersect( Ray& ray, float& dist )  {
        float det = dot( m_normal, ray.get_direction() );
        if( det < 0 ) {
            float distance = -(dot( m_normal, ray.get_origin() ) - m_d) / det;
            if ( distance > 0 && distance < dist) {
                dist = distance;
                return Ray::HIT;
            }
        }
        return Ray::MISS;    
    }
    
    // vector algebra    
    inline void normalize() {
        float inv_length = 1.0f / m_normal.get_length();
        m_normal *= inv_length;
        m_d *= inv_length;
    }
     
    inline bool is_in_front(const Vector3f & point) {
        return distance(point) > 0;
    }    

    inline float distance(const Vector3f & point) {
        return dot(point, m_normal) + m_d;
    }
public:
    // member variables
    Vector3f m_normal;
    float m_d;
};

//==============================================================================
//  Sphere
//==============================================================================
class Sphere
{
public:
    Sphere() {};
    Sphere( Vector3f& centre, float radius ) : m_centre(centre), m_radius(radius) {}
    
    void set_centre( const Vector3f& centre )    { m_centre = centre; }
    void set_radius( float radius )             { m_radius = radius; m_squared_radius = radius*radius; }
    
    Vector3f& get_centre()            { return m_centre; }
    float&    get_radius()            { return m_radius; }
    float     get_squared_radius() { return m_squared_radius; }
         
    int intersect( Ray& ray, float& dist ) {
        Vector3f v =  ray.get_origin() - m_centre;
        float b =  -dot( v, ray.get_direction() );
        float det = (b * b) - dot(v, v) + (m_radius * m_radius);
    
        if (det > 0) {
            det = sqrtf( det );
            float i1 = b - det;
            float i2 = b + det;
        
            if (i2 > 0) {
                if (i1 < 0) {
                    if (i2 < dist) {
                        dist = i2;
                        return Ray::INPRIM;
                    }
                } else {
                    if (i1 < dist) {
                        dist = i1;
                        return Ray::HIT;
                    }
                }
            }
        }
        return Ray::MISS;
    }

    void get_normal( const Vector3f& position, Vector3f& normal) {
        normal = (position - m_centre).get_normalized();
    }    

    inline bool is_inside(const Vector3f & point) {
        return ( (point-m_centre).get_squared_length() <= m_squared_radius);
    }    
protected:
    Vector3f m_centre;
    float m_radius, m_squared_radius;
};

//==============================================================================
//  AA_box
//==============================================================================
class AA_box
{
public:
    AA_box() {}
    AA_box( const Vector3f& centre, const Vector3f& diagonal ) {
		m_diagonal = diagonal;	
        m_centre = centre;
    }
	void set( const Vector3f& centre, const Vector3f& diagonal ) {
		m_diagonal = diagonal;	
        m_centre = centre;
    }    
    void set_min_max( const Vector3f& min, const Vector3f& max ) { 
		m_diagonal = (max - min) * 0.5f;	
        m_centre = (max + min) * 0.5f;
    }
    bool intersect( AA_box* aa_box ) {
		if(fabs(m_centre.x - aa_box->get_centre().x) > (m_diagonal.x + aa_box->get_diagonal().x)) return false;
		if(fabs(m_centre.y - aa_box->get_centre().y) > (m_diagonal.y + aa_box->get_diagonal().y)) return false;
		if(fabs(m_centre.z - aa_box->get_centre().z) > (m_diagonal.z + aa_box->get_diagonal().z)) return false;				
		return true;
	}
    bool intersect( const Vector3f& point ) {
		if(fabs(m_centre.x - point.x) > m_diagonal.x) return false;
		if(fabs(m_centre.y - point.y) > m_diagonal.y) return false;
		if(fabs(m_centre.z - point.z) > m_diagonal.z) return false;				
		return true;
	}	
    int intersect( Ray& ray, float& dist ) {
        Vector3f d = ray.get_direction()*dist*0.5f;       // Segment midpoint
        Vector3f m = d + ray.get_origin();                // Segment halflength vector
        m -= m_centre;                                      // Translate box and segment to origin
    
        // Try world coordinate axes as separating axes
        float adx = fabs(d.x);
        if (fabs(m.x) > m_diagonal.x + adx) return Ray::MISS;
        float ady = fabs(d.y);
        if (fabs(m.y) > m_diagonal.y + ady) return Ray::MISS;
        float adz = fabs(d.z);
        if (fabs(m.z) > m_diagonal.z + adz) return Ray::MISS;
    
        // Add in an epsilon term to counteract arithmetic errors when segment is
        // (near) parallel to a coordinate axis (see text for detail)
//        adx += EPSILON; ady += EPSILON; adz += EPSILON;
    
        // Try cross products of segment direction vector with coordinate axes
        if (fabs(m.y * d.z - m.z * d.y) > m_diagonal.y * adz + m_diagonal.z * ady) return Ray::MISS;
        if (fabs(m.z * d.x - m.x * d.z) > m_diagonal.x * adz + m_diagonal.z * adx) return Ray::MISS;
        if (fabs(m.x * d.y - m.y * d.x) > m_diagonal.x * ady + m_diagonal.y * adx) return Ray::MISS;
    
        // No separating axis found; segment must be overlapping AABB
        return Ray::HIT;
    }
	const Vector3f &get_centre()   { return m_centre;   }
	const Vector3f &get_diagonal() { return m_diagonal; }	
protected:
    Vector3f m_centre, m_diagonal;
};

//==============================================================================
//  Triangle
//==============================================================================
class Triangle
{
public:
    Triangle() {}
    Triangle( const Vector3f& v1, const Vector3f& v2, const Vector3f& v3 ) : 
                m_v1(v1), m_v2(v2), m_v3(v3) {	
        ab = m_v2 - m_v1;
        ac = m_v3 - m_v1;      
        n = cross(ab, ac);					  
    }
                
//    void set_normal( Vector3f normal )        { m_normal = normal; }
//    const Vector3f& get_normal()              { return m_normal; }
        
    inline int intersect( Ray& ray, float& dist ) {
        Vector3f qp = -ray.get_direction() * dist; //p - q;
            
        float d = dot(qp, n );
        
        if (d <= 0.0f) {
               d = -d;  
            
            // Compute intersection t value of pq with plane of triangle. A ray
            // intersects iff 0 <= t. Segment intersects iff 0 <= t <= 1. Delay
            // dividing by d until intersection has been found to pierce triangle
            Vector3f ap = ray.get_origin()- m_v1;
            float t = dot(ap, -n);
            if (t < 0.0f) return Ray::MISS;
            if (t > d) return Ray::MISS; // For segment; exclude this code line for a ray test
    
            // Compute barycentric coordinate components and test if within bounds
            Vector3f e = cross(qp, ap);
            float v = dot(ab, e);
            if (v < 0.0f || v > d) return Ray::MISS;
            float w = -dot(ac, e);
            if (w < 0.0f || v + w > d) return Ray::MISS;
    
            // Segment/ray intersects triangle. Perform delayed division and
            // compute the last barycentric coordinate component
            ood = 1.0f / d;
            t *= ood;            
            dist *= t;
        	m_v = w;
        	m_u = v;
        
            return dot(m_normal_1, qp)<=0? Ray::INPRIM : Ray::HIT;
        } else {                   
            // Compute intersection t value of pq with plane of triangle. A ray
            // intersects iff 0 <= t. Segment intersects iff 0 <= t <= 1. Delay
            // dividing by d until intersection has been found to pierce triangle
            Vector3f ap = ray.get_origin()- m_v1;
            float t = dot(ap, n);
            if (t < 0.0f) return Ray::MISS;
            if (t > d) return Ray::MISS; // For segment; exclude this code line for a ray test
    
            // Compute barycentric coordinate components and test if within bounds
            Vector3f e =cross(qp, ap);
            float v = dot(ac, e);
            if (v < 0.0f || v > d) return Ray::MISS;
            float w = -dot(ab, e);
            if (w < 0.0f || v + w > d) return Ray::MISS;
    
            // Segment/ray intersects triangle. Perform delayed division and
            // compute the last barycentric coordinate component
            ood = 1.0f / d;
            t *= ood;
            dist *= t;
        	m_v = v;
        	m_u = w;
        
            return dot(m_normal_1, qp)<=0? Ray::INPRIM : Ray::HIT;
        }
    }

public:
    Vector3f m_v1, m_v2, m_v3;
    Vector3f ab, ac, n;
    float m_v, m_u, ood;
    Vector3f m_normal_1, m_normal_21, m_normal_31;
    Vector2f m_texcoord_1, m_texcoord_21, m_texcoord_31;
    Vector3f m_tangent_s_1, m_tangent_s_21, m_tangent_s_31;
};


#endif    /* _GEOMETRY_ */

