//============================================================================//
// This source file is part of work done by Reinder Nijhoff (reinder@infi.nl) //
// For the latest info, see http://developer.infi.nl                          //
//                                                                            //
// You're free to use the code in any way you like, modified, unmodified or   //
// cut'n'pasted into your own work.                                           //
//                                                                            //
// Part of this source is based on work by:                                   //
//    - Humus (http://esprit.campus.luth.se/~humus/)                          //
//    - Paul Baker (http://www.paulsprojects.net)                             //
//============================================================================//
#include <string.h>
#include "image.hpp"
#include "log.hpp"
#include "math.hpp"
#include "factory.hpp"

#include <fstream>

static int TGAReadError = 0;
   
//==============================================================================
//  Register to Factory
//==============================================================================
bool fac_image  = Factory_image.reg<Image>( "image" );

//==============================================================================
//  read_data()
//==============================================================================
void read_data(std::ifstream &file, char* data, unsigned int size) {
    if (!file.is_open()) {
        return;
    }    
    unsigned int a = file.tellg();
    a+= size;
    file.read(data, size);
    if (a != file.tellg()) {
        TGAReadError = 1;
    }
}

//==============================================================================
//  Image::Image()
//==============================================================================
Image::Image() {
    m_pixels = NULL;
	free_data();			   
}

//==============================================================================
//  Image::Image()
//==============================================================================
Image::Image(const std::string &filename) {
    Image();
    load_from_file(filename);
}

//==============================================================================
//  Image::~Image()
//==============================================================================
Image::~Image() {
    free_data();
}

//==============================================================================
//  Image::load_from_file_pfm()
//==============================================================================
bool Image::load_from_file_pfm(const std::string &filename) {
    log_init_multiple << "Load " << filename << " in Image::load_from_file()" << std::endl;
    free_data();
    
    std::ifstream file;
    file.open(filename.c_str(), std::ios::binary);
 
    if (!file.is_open()) {
	    create_empty();
        return false;
    }
	
    int width, height;
	char buffer[256];
	
	file.getline(buffer, 256);
	file.getline(buffer, 256);
	
	if ( sscanf(buffer, "%i %i", &width, &height) != 2  ) {
		file.close();
	    create_empty();
		return false;
	}
	
	file.getline(buffer, 256);

	create_empty(width, height, Image::RGB, Image::FLOAT);
	
    file.read( (char *)m_pixels, width * height * 3 * sizeof(float) );
    	
	file.close();
	return true;
}
    
//==============================================================================
//  Image::load_from_file()
//==============================================================================
bool Image::load_from_file(const std::string &filename) {
    log_init << "Load " << filename << " in Image::load_from_file()" << std::endl;
    
    free_data();
    m_format = Image::CHAR;
    
    std::ifstream file;
    file.open(filename.c_str(), std::ios::binary);
    if (!file.is_open()) {
	    create_empty();
        fill_as_default();
        return false;
    }
    
    bool rle = false;
    bool truecolor = false;
    unsigned int Currentpixel = 0;
    unsigned char ch_buf1, ch_buf2;
    unsigned char buf1[1000];

    unsigned char IDLength;
    unsigned char IDColorMapType;
    unsigned char IDImaget_ype;

    read_data(file, (char*)&IDLength, 1);
    read_data(file, (char*)&IDColorMapType, 1);
    
    if (IDColorMapType == 1) {
        create_empty();
        fill_as_default();
        return false;
    }
    
    read_data(file, (char*)&IDImaget_ype, 1); 

    switch (IDImaget_ype) {
    case 2:
            truecolor = true;
            break;
    case 3:
            m_type = Image::GREYSCALE;
            break;
    case 10:
            rle = true;
            truecolor = true;
            break;
    case 11:
            rle = true;
            m_type = Image::GREYSCALE;
            break;
    default:
            return false;
    }

    file.seekg(5, std::ios::cur);

    file.seekg(4, std::ios::cur);
    read_data(file, (char*)&m_width, 2);
    read_data(file, (char*)&m_height, 2);
    read_data(file, (char*)&m_pixel_depth, 1);

    if (! ((m_pixel_depth == 8) || (m_pixel_depth ==  24) ||
           (m_pixel_depth == 16) || (m_pixel_depth == 32))) {   
        create_empty();   
        fill_as_default();           
        return false;
    }    

    read_data(file, (char*)&ch_buf1, 1); 
    
    ch_buf2 = 15; //00001111;
    m_alpha_depth = ch_buf1 & ch_buf2;

    if (! ((m_alpha_depth == 0) || (m_alpha_depth == 8))) {
        create_empty();
        return false;
    }
    
    if (truecolor) {
        m_type = Image::RGB;
        if (m_pixel_depth == 32) {
            m_type = Image::RGBA;
        }    
    }    

    if (m_type == Image::UNDEFINED) {
        create_empty();
        fill_as_default();
        return false;
    }
    
    file.seekg(IDLength, std::ios::cur);

    m_pixels = new unsigned char[m_width*m_height*(m_pixel_depth/8)];

    if (!rle) {
        read_data(file, (char*)m_pixels, m_width*m_height*(m_pixel_depth/8));
    } else {
        while (Currentpixel < m_width*m_height-1) {
            
            read_data(file, (char*)&ch_buf1, 1);
            if ((ch_buf1 & 128) == 128) {   // this is an rle packet
                ch_buf2 = (unsigned char)((ch_buf1 & 127) + 1);   // how many pixels are encoded using this packet
                read_data(file, (char*)buf1, m_pixel_depth/8);
                for (unsigned int i=Currentpixel; i<Currentpixel+ch_buf2; ++i) {
                    for (unsigned int j=0; j<m_pixel_depth/8; ++j) {
                        m_pixels[i*m_pixel_depth/8+j] = buf1[j];
                    }
                }        
                Currentpixel += ch_buf2;
            } else {   // this is a raw packet
                ch_buf2 = (unsigned char)((ch_buf1 & 127) + 1);
                read_data(file, (char*)buf1, m_pixel_depth/8*ch_buf2);
                for (unsigned int i=Currentpixel; i<Currentpixel+ch_buf2; ++i) {
                    for (unsigned int j=0; j<m_pixel_depth/8; ++j) {
                        m_pixels[i*m_pixel_depth/8+j] =  buf1[(i-Currentpixel)*m_pixel_depth/8+j];
                    }    
                }    
                Currentpixel += ch_buf2;
            }
        }
    }

    if (TGAReadError != 0) {
        create_empty();
        fill_as_default();
        return false;
    }

    // swap BGR(A) to RGB(A)

    unsigned char temp;
    if ((m_type == Image::RGB) || (m_type == Image::RGBA)) {
        if ((m_pixel_depth == 24) || (m_pixel_depth == 32)) {
            for (unsigned int i=0; i<m_width*m_height; ++i)
            {
                temp = m_pixels[i*m_pixel_depth/8];
                m_pixels[i*m_pixel_depth/8] = m_pixels[i*m_pixel_depth/8+2];
                m_pixels[i*m_pixel_depth/8+2] = temp;
            }
        }
    }

    m_width_minus_one = m_width - 1;
    m_height_minus_one = m_height - 1;
	if( m_type == Image::RGB ) {
		m_pixel_size = 3;
	} else {
		m_pixel_size = 4;
    }

    flip_vertically( this );
//	change_format( Image::FLOAT );
    return true;
}

//==============================================================================
//  Image::free_data()
//==============================================================================
void Image::free_data() {
    if( m_pixels != NULL ) {
        delete[] m_pixels;
    }
    
    m_width = 0;
    m_height = 0;
    m_pixel_depth = 0;
    m_alpha_depth = 0;
    m_type = Image::UNDEFINED;
    m_format = Image::CHAR;
    m_pixels = NULL;
}

//==============================================================================
//  Image.create_empty()
//==============================================================================
void Image::create_empty() {
    create_empty(64, 64, Image::RGB, Image::CHAR );
}

//==============================================================================
//  Image.create_empty()
//==============================================================================
void Image::create_empty(int width, int height, Image_type type, Image_format format) {
    free_data();
    
    // create a default texture    
    m_width = width;
    m_height = height;
    m_type = type;
    m_format = format;
    m_alpha_depth = 0;
    m_width_minus_one = m_width - 1;
    m_height_minus_one = m_height - 1;
        
    switch(type) {
        case Image::RGB:  m_pixel_depth = 24;
			 			  m_pixel_size = 3;
                          break;
        case Image::RGBA: m_pixel_depth = 32;
                          m_alpha_depth = 8;
			 			  m_pixel_size = 4;
                          break;        
    }    

    if( m_format == Image::FLOAT ) {
        m_pixel_depth *= sizeof(float)/sizeof(char);
    }
    
    m_pixels = new unsigned char[m_width*m_height*m_pixel_depth/sizeof(char)];
}

//==============================================================================
//  Image::fill_as_default()
//==============================================================================
void Image::fill_as_default() {
    fill(orange);
    
    // top line
    for(int i=0; i<m_width; ++i) {
        put_pixel(i, 0, white);
       }       
    // left line
    for(int i=0; i<m_height; ++i) {
        put_pixel(0, i, white);
    }
}

//==============================================================================
//  Image::fill()
//==============================================================================
void Image::fill(const Color& color) {    
    for( int x=0; x<m_width; ++x ) {
        for( int y=0; y<m_height; ++y ) {
            put_pixel( x, y, color );
        }
    }
}

//==============================================================================
//  Image::clear()
//==============================================================================
void Image::clear() {
     memset( m_pixels, 0, m_width*m_height*m_pixel_depth/sizeof(char) );
}

//==============================================================================
//  Image::apply_exposure()
//==============================================================================
void Image::apply_exposure(float exp) {
     for(int x=0; x<m_width; ++x) {
        for(int y=0; y<m_height; ++y) {
            put_pixel( x, y, exposure( get_pixel( x, y ), exp ) );
        }    
     }     
}

//==============================================================================
//  Image::change_format()
//==============================================================================
void Image::change_format( Image_format format ) {
	if( format == m_format ) return;
	
	if( format == Image::FLOAT ) {
		unsigned int pixel_depth = 24; //m_pixel_depth * sizeof(float)/sizeof(char);
		log_debug << pixel_depth << std::endl;
        unsigned char* pixels = new unsigned char[m_width*m_height*pixel_depth/sizeof(char)];
        float val;
        
        for(int offset=0; offset<m_width * m_height * m_pixel_size; ++offset) {
		    ((float *)pixels)[ offset ] = (float)(m_pixels[ offset ]) / 255.0f;	
		}
		delete[] m_pixels;
		m_pixels = pixels;
		m_format = format;
		m_pixel_depth = pixel_depth;
	} 
}

//==============================================================================
//  flip_vertically()
//==============================================================================
void flip_vertically(Image* image) {
	unsigned char* data = image->get_pixels();
	unsigned int height = image->get_height();
	unsigned int width = image->get_width();
	unsigned int bpp = image->get_pixel_depth();

	if( height==0 || height==1) {
		return;
    }
	int rows_to_swap = 0;
	
	if(height % 2 == 1) {
		rows_to_swap = (height-1)/2;
    } else {
		rows_to_swap = height/2;
    }     

	unsigned char * temp_row=new unsigned char[width * bpp/8];

	for(int i=0; i<rows_to_swap; ++i) {
		memcpy(temp_row, &data[i*width*bpp/8], width*bpp/8);
		memcpy(&data[i*width*bpp/8], &data[(height-i-1)*width*bpp/8], width*bpp/8);
		memcpy(&data[(height-i-1)*width*bpp/8], temp_row, width*bpp/8);
	}

    delete [] temp_row;
}

//==============================================================================
//  get_image_from_file()
//==============================================================================
Image* get_image_from_file( const std::string& name ) {	
	if( Image* image = Factory_image.use(name) ) {
		return image;
    } else {
	  	Image* image = Factory_image.create("image", name);
		image->load_from_file( name );
		return image;
	}   
}
