//============================================================================//
// This source file is part of work done by Reinder Nijhoff (reinder@infi.nl) //
// For the latest info, see http://developer.infi.nl                          //
//                                                                            //
// You're free to use the code in any way you like, modified, unmodified or   //
// cut'n'pasted into your own work.                                           //
//                                                                            //
// Part of this source is based on work by:                                   //
//    - Humus (http://esprit.campus.luth.se/~humus/)                          //
//    - Paul Baker (http://www.paulsprojects.net)                             //
//============================================================================//

#ifndef  _IMAGE_
#define  _IMAGE_ 1

#include <string>

#include "includes.hpp"
#include "vector.hpp"

//==============================================================================
//  Image
//==============================================================================
class Image
{
public:
    Image();
    Image(const std::string &filename);
    ~Image();
    
    enum Image_type {UNDEFINED, RGB, RGBA, GREYSCALE};
    enum Image_format {FLOAT, CHAR};
        
    bool load_from_file(const std::string &filename);
    bool load_from_file_pfm(const std::string &filename);
    
    void free_data();
    void clear();
    void create_empty();
    void create_empty(int width, int height, Image_type type, Image_format format);
    
    void change_format( Image_format format );
    
    inline unsigned int get_alpha_depth() { return m_alpha_depth; }
    inline unsigned int get_pixel_depth() { return m_pixel_depth; }
    
    inline unsigned int get_width()       { return m_width; }
    inline unsigned int get_height()      { return m_height; }
    inline unsigned int get_width_mask()  { return m_width_minus_one; }
    inline unsigned int get_height_mask() { return m_height_minus_one; }

    unsigned char* get_pixels()           { return m_pixels; }
    Image::Image_type get_image_type()    { return m_type; }
        
    void fill(const Color& color);
    void fill_as_default();
    
    void apply_exposure(float exp);
    
//==============================================================================
//  Image::put_pixel()
//==============================================================================
    void put_pixel(int x, int y, const Color& color) {
//        if(x<0 || y<0 || x>=m_width || y >=m_height) return;
        int offset = (x+(y*m_width)) * m_pixel_size;
        
        if( m_format == Image::FLOAT ) {
            ((float *) m_pixels)[ offset+0 ] = color.x;
            ((float *) m_pixels)[ offset+1 ] = color.y;
            ((float *) m_pixels)[ offset+2 ] = color.z;                
        } else {
            m_pixels[ offset+0 ] = (unsigned char)(color.x*255.0f);
            m_pixels[ offset+1 ] = (unsigned char)(color.y*255.0f);
            m_pixels[ offset+2 ] = (unsigned char)(color.z*255.0f);
        }
    }

//==============================================================================
//  Image::get_pixel()
//==============================================================================
    inline Color get_pixel(int x, int y) {        
//        if(x<0 || y<0 || x>=m_width || y >=m_height) return black;
        
        int offset = (x+(y*m_width)) * m_pixel_size;
        Color color;
			            
        if( m_format == Image::FLOAT ) {
            color.x = ((float *)m_pixels)[ offset+0 ];
            color.y = ((float *)m_pixels)[ offset+1 ];
            color.z = ((float *)m_pixels)[ offset+2 ];
       } else {
            color.x = (float)(m_pixels[ offset+0 ]);// / 255.0f;
            color.y = (float)(m_pixels[ offset+1 ]);// / 255.0f;
            color.z = (float)(m_pixels[ offset+2 ]);// / 255.0f;
            color *=  0.00392f; // * 1.0f/255.0f;
        }
        
        return color;
    }

//==============================================================================
//  Image::get_float()
//==============================================================================
    inline float get_float(int x, int y, int index=0) {                
        int offset = (x+(y*m_width)) * m_pixel_size + index;
        Color color;
			            
        if( m_format == Image::FLOAT ) {
            return ((float *)m_pixels)[ offset ];
        } else {
            return (float)(m_pixels[ offset+0 ]) * 0.00392f;// / 255.0f;
        }
    }
   
//==============================================================================
//  Image::add_pixel()
//==============================================================================
    void add_pixel(int x, int y, const Color& color) {
        Color new_color = get_pixel( x, y ) + color;
        put_pixel( x, y, new_color );
    }

//==============================================================================
//  Image::get_pixel()
//==============================================================================  
template <class T>
    inline T get_value(int x, int y);

template <class T>
    inline T get_pixel( float u, float v, bool clamp ) {		   
        u *= (float) get_width();
        v *= (float) get_height();
        
        int u_i = (int)u;
        int v_i = (int)v;
        
        int u_ix = u_i + 1;
        int v_ix = v_i + 1;
        
        float u_frac = u - (float)u_i;
        float v_frac = v - (float)v_i;
        if( u_frac < 0.0f ) u_frac += 1.0f;
        if( v_frac < 0.0f ) v_frac += 1.0f;

        const int width_mask = get_width_mask();
        const int height_mask = get_height_mask();
    
        if( !clamp ) {        
            u_ix &= width_mask;
            u_i &= width_mask;
            
            v_ix &= height_mask;
            v_i &= height_mask;
        } else {
            u_i = get_max( 0, get_min( u_i, width_mask ) );            
            u_ix = get_max( 0, get_min( u_ix, width_mask ) );
            
            v_i = get_max( 0, get_min( v_i, height_mask ) );
            v_ix = get_max( 0, get_min( v_ix, height_mask ) );
        }

		const int m_u_i = u_i, m_u_ix = u_ix;
		const int m_v_i = v_i, m_v_ix = v_ix;   
		const float m_u_frac = u_frac, m_v_frac = v_frac;
		
	    return lerp( lerp( get_value<T>( m_u_i, m_v_i ),  get_value<T>( m_u_ix, m_v_i ),  m_u_frac ),
               lerp( get_value<T>( m_u_i, m_v_ix ), get_value<T>( m_u_ix, m_v_ix ), m_u_frac ), m_v_frac);
    }
    
protected:
    unsigned char *m_pixels, m_pixel_size;
    unsigned int m_pixel_depth, m_alpha_depth;
    unsigned int m_height, m_width;
    unsigned int m_height_minus_one, m_width_minus_one;
    
    Image_type m_type;
    Image_format m_format;
};

template <>
inline float Image::get_value(int x, int y) { return get_float( x, y ); }

template <>
inline Color Image::get_value(int x, int y) { return get_pixel( x, y ); }    


//==============================================================================
//  flip_vertically()
//==============================================================================
void flip_vertically(Image* image);

//==============================================================================
//  get_image_from_file()
//==============================================================================
Image* get_image_from_file( const std::string& name );

#endif /* _IMAGE_ */

