//============================================================================//
// This source file is part of work done by Reinder Nijhoff (reinder@infi.nl) //
// For the latest info, see http://developer.infi.nl                          //
//                                                                            //
// You're free to use the code in any way you like, modified, unmodified or   //
// cut'n'pasted into your own work.                                           //
//                                                                            //
// Part of this source is based on work by:                                   //
//    - Humus (http://esprit.campus.luth.se/~humus/)                          //
//    - Paul Baker (http://www.paulsprojects.net)                             //
//============================================================================//

#ifndef _RENDERER_
#define _RENDERER_ 1

#include <string>
#include <map>

#include "includes.hpp"
#include "vector.hpp"
#include "matrix.hpp"
#include "data.hpp"

class Batch;
class Texture;
class Shader;
class Camera;
class Material;
class Primitive;
class Background;
class AA_box;

#define TEXTURE_NONE  (-1)
#define SHADER_NONE   NULL
#define DEPTH_NONE    (-1)
#define BLEND_NONE (-1)
#define CULL_NONE (-1)

// Mask constants
extern int BUFFER_DEPTH;
extern int BUFFER_COLOR;

// Blending constants
extern int BLEND_ZERO;
extern int BLEND_ONE;
extern int BLEND_SRC_COLOR;
extern int BLEND_ONE_MINUS_SRC_COLOR;
extern int BLEND_DST_COLOR;
extern int BLEND_ONE_MINUS_DST_COLOR;
extern int BLEND_SRC_ALPHA;
extern int BLEND_ONE_MINUS_SRC_ALPHA;
extern int BLEND_DST_ALPHA;
extern int BLEND_ONE_MINUS_DST_ALPHA;
extern int BLEND_SRC_ALPHA_SATURATE;

extern int BLEND_ADD;
extern int BLEND_SUBTRACT;
extern int BLEND_REVERSE_SUBTRACT;
extern int BLEND_MIN;
extern int BLEND_MAX;

// Depth testing constants
extern int DEPTH_NEVER;
extern int DEPTH_LESS;
extern int DEPTH_EQUAL;
extern int DEPTH_LEQUAL;
extern int DEPTH_GREATER;
extern int DEPTH_NOTEQUAL;
extern int DEPTH_GEQUAL;
extern int DEPTH_ALWAYS;

// Culling constants
extern int CULL_BACK;
extern int CULL_FRONT;


//==============================================================================
//  Renderer
//==============================================================================
class Renderer {
public:
    Renderer();
    virtual ~Renderer() {}
    
    virtual void init(Data_node* data);
    
    // init    
    virtual void init()=0;
    virtual void close()=0;
    virtual void start_renderer();
    virtual void stop_renderer();
    virtual bool get_hardware_acc() { return true; }
    
    // scene
    virtual void render_scene( std::vector<Primitive *> *primitives, std::vector<Primitive *> *lights, Background* background = NULL) {}

    // viewport
    virtual void set_viewport(const int left, const int top, const int width, const int height)=0;
    virtual void set_renderport(const int left, const int top, const int width, const int height)=0;
    virtual void set_view(Camera* camera)=0;
    virtual void set_view_rotation(Camera* camera)=0;
    virtual void clear_buffer( int mask ) {}
    
	// 2d
	virtual void enable_2d_mode( int width, int height ) {}    
	virtual void disable_2d_mode() {}    
    
    // texture
    virtual void bind_texture(int texture_slot, Texture* texture)=0;
    virtual void activate_texture_slot(int texture_slot, int tex_id) {}
    virtual void disable_texture_slot(int texture_slot) {}
    virtual void disable_texture_slots() {}

    // shader
    virtual void set_shader(Shader* shader) {}

    // camera
    inline Camera* get_camera() { return m_current_camera; }
    void set_camera(Camera* camera) { m_current_camera = camera; }    

	// misc functions	
	virtual void set_cull_face(const int cull_face) {}
	virtual void set_depth_func(const int depth_func) {}
	virtual void set_mask(const int mask) {}
	virtual void set_blending(const int src, const int dest, const int mode = BLEND_ADD) {}
    virtual void set_scissor(const int x, const int y, const int w, const int h) {}
    virtual void set_color( const Color& color ) {}
    virtual void set_default() {}
    
    // frames
    inline int get_frame() { return m_frame; }
    void next_frame() { m_frame++; }    
    
    // debug functions 
    void render_debug();    
	void add_2d_debug_line( Vector2f start, Vector2f end );
	void add_2d_debug_quad( Vector2f tl, Vector2f br );
	
    void add_3d_debug_line( Vector3f start, Vector3f end );    
    void add_3d_debug_aa_box( Vector3f tlf, Vector3f brb );
    void add_3d_debug_aa_box( AA_box* aa_box );
	
	virtual void render_debug_lines();
protected:
    int m_frame;
    float m_starttime, m_totaltime;
            
    // camera
    Camera *m_current_camera, *m_default_camera;
    
    // debug
    std::vector< Vector3f > m_debug_lines_3d;
    std::vector< Vector2f > m_debug_lines_2d;
        
	// misc functions	
	unsigned int m_shader_index;
	int m_cull_face, m_depth_func, m_mask;
	int m_blend_src, m_blend_dest, m_blend_mode;
	
private:
};    

extern Renderer* renderer;

#endif /* _RENDERER_ */
