//============================================================================//
// This source file is part of work done by Reinder Nijhoff (reinder@infi.nl) //
// For the latest info, see http://developer.infi.nl                          //
//                                                                            //
// You're free to use the code in any way you like, modified, unmodified or   //
// cut'n'pasted into your own work.                                           //
//                                                                            //
// Part of this source is based on work by:                                   //
//    - Humus (http://esprit.campus.luth.se/~humus/)                          //
//    - Paul Baker (http://www.paulsprojects.net)                             //
//============================================================================//

#include "scene.hpp"
#include "background.hpp"
#include "primitive.hpp"
#include "camera.hpp"
#include "batch.hpp"

#include "log.hpp"
#include "data.hpp"
#include "renderer.hpp"
#include "base.hpp"
#include "factory.hpp"
#include "material.hpp"
#include "image.hpp"
#include "texture.hpp"
#include "shader.hpp"

//==============================================================================
//  Register to Factory
//==============================================================================
bool fac_scene = Factory_scene.reg< Scene >( "scene" );

//==============================================================================
//  Scene::Scene()
//==============================================================================
Scene::Scene() {
	m_background = NULL;
}

//==============================================================================
//  Scene::~Scene()
//==============================================================================
Scene::~Scene() {
    free_data();
}

//==============================================================================
//  Scene::free_data()
//==============================================================================
void Scene::free_data() {
	m_background = NULL;
	m_primitives.clear();
	m_cameras.clear();
	m_lights.clear();
	m_batches.clear();
}

//==============================================================================
//  Scene::init()
//==============================================================================
void Scene::init(Data_node* data) {
    std::string s;    
    
    free_data();
        
    data->get_value( "log_level", logger.get_log_level() );
    data->reset();
    
    while( Data_value* value = data->get_next_value() ) {
	    s = value->get_key();
	    		    
        if( s == "texture" ) {     
            Factory_texture.create( value );
        } else if( s == "material" ) {
            Factory_material.create( value );
        } else if( s == "primitive" ) {
            Primitive* primitive = Factory_primitive.create( value );
            primitive->first_update();
            m_primitives.push_back( primitive );
        } else if( s == "camera" ) {
            Camera* camera = Factory_camera.create( value );
            camera->first_update();
            m_cameras.push_back( camera );
        } else if( s == "light" ) {
            Primitive* light = Factory_primitive.create( value );
            light->set_type( Primitive::LIGHT );
            light->first_update();            
            m_lights.push_back( light );
        } else if( s == "background" ) {
            m_background = Factory_background.create( value );
        } else if( s == "batch" ) {
            Batch* batch = Factory_batch.create( value );
            m_batches.push_back( batch );
        } else if( s == "shader" ) {
		  	Shader* shader = Factory_shader.create( value );   
		  	shader->first_update();
		} else if( s == "base" || s == "renderer") {
		}
    }
}

//==============================================================================
//  Scene::render()
//==============================================================================
void Scene::render() {
    renderer->clear_buffer( BUFFER_COLOR | BUFFER_DEPTH );
    renderer->render_scene( &m_primitives, &m_lights, m_background );
}

//==============================================================================
//  Scene::update()
//==============================================================================
bool Scene::update() {
	for(int i=0, j=m_primitives.size(); i<j; ++i) {
		m_primitives[i]->update();	
    }
	for(int i=0, j=m_lights.size(); i<j; ++i) {
		m_lights[i]->update();	
    }    
    renderer->get_camera()->update();
    
    return true;
}

//==============================================================================
//  Scene::first_update()
//==============================================================================
void Scene::first_update() {
	for(int i=0, j=m_primitives.size(); i<j; ++i) {
		m_primitives[i]->first_update();
		m_primitives[i]->calculate_bounding_box();
    }
	for(int i=0, j=m_lights.size(); i<j; ++i) {
		m_lights[i]->first_update();
		m_lights[i]->calculate_bounding_box();
    }    
}
